@@ -1,101 +1,90 @@
 # Copyright (c) OpenMMLab. All rights reserved.
 import torch
-from torch import nn
+from torch import Tensor, nn
 from torch.autograd import Function
 
-from ..utils import ext_loader
-
-ext_module = ext_loader.load_ext('_ext', [
-    'top_pool_forward', 'top_pool_backward', 'bottom_pool_forward',
-    'bottom_pool_backward', 'left_pool_forward', 'left_pool_backward',
-    'right_pool_forward', 'right_pool_backward'
-])
-
 _mode_dict = {'top': 0, 'bottom': 1, 'left': 2, 'right': 3}
 
 
+def _corner_pool(x: Tensor, dim: int, flip: bool) -> Tensor:
+    size = x.size(dim)
+    output = x.clone()
+
+    ind = 1
+    while ind < size:
+        if flip:
+            cur_start = 0
+            cur_len = size - ind
+            next_start = ind
+            next_len = size - ind
+        else:
+            cur_start = ind
+            cur_len = size - ind
+            next_start = 0
+            next_len = size - ind
+
+        # max_temp should be cloned for backward computation
+        max_temp = output.narrow(dim, cur_start, cur_len).clone()
+        cur_temp = output.narrow(dim, cur_start, cur_len)
+        next_temp = output.narrow(dim, next_start, next_len)
+
+        cur_temp[...] = torch.where(max_temp > next_temp, max_temp, next_temp)
+
+        ind = ind << 1
+
+    return output
+
+
 class TopPoolFunction(Function):
 
     @staticmethod
-    def symbolic(g, input):
+    def symbolic(g, input: Tensor) -> Tensor:
         output = g.op(
             'mmcv::MMCVCornerPool', input, mode_i=int(_mode_dict['top']))
         return output
 
     @staticmethod
-    def forward(ctx, input):
-        output = ext_module.top_pool_forward(input)
-        ctx.save_for_backward(input)
-        return output
-
-    @staticmethod
-    def backward(ctx, grad_output):
-        input, = ctx.saved_tensors
-        output = ext_module.top_pool_backward(input, grad_output)
-        return output
+    def forward(ctx, input: Tensor) -> Tensor:
+        return _corner_pool(input, 2, True)
 
 
 class BottomPoolFunction(Function):
 
     @staticmethod
-    def symbolic(g, input):
+    def symbolic(g, input: Tensor) -> Tensor:
         output = g.op(
             'mmcv::MMCVCornerPool', input, mode_i=int(_mode_dict['bottom']))
         return output
 
     @staticmethod
-    def forward(ctx, input):
-        output = ext_module.bottom_pool_forward(input)
-        ctx.save_for_backward(input)
-        return output
-
-    @staticmethod
-    def backward(ctx, grad_output):
-        input, = ctx.saved_tensors
-        output = ext_module.bottom_pool_backward(input, grad_output)
-        return output
+    def forward(ctx, input: Tensor) -> Tensor:
+        return _corner_pool(input, 2, False)
 
 
 class LeftPoolFunction(Function):
 
     @staticmethod
-    def symbolic(g, input):
+    def symbolic(g, input: Tensor) -> Tensor:
         output = g.op(
             'mmcv::MMCVCornerPool', input, mode_i=int(_mode_dict['left']))
         return output
 
     @staticmethod
-    def forward(ctx, input):
-        output = ext_module.left_pool_forward(input)
-        ctx.save_for_backward(input)
-        return output
-
-    @staticmethod
-    def backward(ctx, grad_output):
-        input, = ctx.saved_tensors
-        output = ext_module.left_pool_backward(input, grad_output)
-        return output
+    def forward(ctx, input: Tensor) -> Tensor:
+        return _corner_pool(input, 3, True)
 
 
 class RightPoolFunction(Function):
 
     @staticmethod
-    def symbolic(g, input):
+    def symbolic(g, input: Tensor) -> Tensor:
         output = g.op(
             'mmcv::MMCVCornerPool', input, mode_i=int(_mode_dict['right']))
         return output
 
     @staticmethod
-    def forward(ctx, input):
-        output = ext_module.right_pool_forward(input)
-        ctx.save_for_backward(input)
-        return output
-
-    @staticmethod
-    def backward(ctx, grad_output):
-        input, = ctx.saved_tensors
-        output = ext_module.right_pool_backward(input, grad_output)
-        return output
+    def forward(ctx, input: Tensor) -> Tensor:
+        return _corner_pool(input, 3, False)
 
 
 class CornerPool(nn.Module):
@@ -135,13 +124,13 @@ class CornerPool(nn.Module):
         'top': (2, True),
     }
 
-    def __init__(self, mode):
-        super(CornerPool, self).__init__()
+    def __init__(self, mode: str):
+        super().__init__()
         assert mode in self.pool_functions
         self.mode = mode
-        self.corner_pool = self.pool_functions[mode]
+        self.corner_pool: Function = self.pool_functions[mode]
 
-    def forward(self, x):
+    def forward(self, x: Tensor) -> Tensor:
         if torch.__version__ != 'parrots' and torch.__version__ >= '1.5.0':
             if torch.onnx.is_in_onnx_export():
                 assert torch.__version__ >= '1.7.0', \
@@ -160,4 +149,8 @@ class CornerPool(nn.Module):
                 pool_tensor = pool_tensor.flip(dim)
             return pool_tensor
         else:
-            return self.corner_pool.apply(x)
+            if torch.onnx.is_in_onnx_export():
+                return self.corner_pool.apply(x)
+            else:
+                dim, flip = self.cummax_dim_flip[self.mode]
+                return _corner_pool(x, dim, flip)
